import logging
import os
import sys
import argparse

import warnings
warnings.simplefilter(action='ignore', category=UserWarning)
import shapely
import geopandas as gpd

try:
    from sites import create_sites_file
except ModuleNotFoundError:
    for path in sys.path:
        if "prebas_weather" in os.listdir(path):
            sys.path.append(os.path.join(path, "prebas_weather"))
            from sites import create_sites_file
            break

from site_processor import get_data_from_clipick
logging.getLogger().setLevel(logging.INFO)

clipick_extent_wkt = "POLYGON ((23.47841578340817 34.39788005192901, 14.25327875377249 35.33928230717498, 13.13483184162019 36.22363567957447, 11.05400037715079 36.14560449965687, -6.893171003897777 34.66301208122243, -9.130064828202382 35.13119916072804, -24.93076907914224 63.08858354916435, -25.45991436426534 66.02827957762601, -24.34282987344992 66.96898230673375, -20.52122503644976 66.85139446559528, -19.93328583075743 67.96847895641071, -16.40565059660344 67.20415798901068, 6.999470176817231 65.6193153140163, 14.64163755884682 71.28506009724514, 28.67423801033218 71.94386763017873, 43.29976524145776 69.04511448527094, 45.27618784025852 69.11099523856431, 50.61252885702055 66.40988435353661, 40.31397457104929 58.02929279476642, 37.39889048539511 54.33163580021179, 35.26637930864139 51.35785874639009, 32.39203249604414 46.46751416521121, 30.25958830641831 41.47759476148677, 30.13164165504076 41.0511059235616, 29.06541956022784 36.65827089293239, 28.29773965196255 35.50675103053445, 26.84767760301698 34.39788005192901, 23.47841578340817 34.39788005192901))"

def get_weather_data(aoi_path, sites_path, weather_path, start_date, end_date, step=25):

    clipick_extent = shapely.wkt.loads(clipick_extent_wkt) 
    clipick_extent_gdf = gpd.GeoDataFrame({"id" : [1], "geometry": [clipick_extent]}, crs="EPSG:4326")

    aoi = gpd.read_file(aoi_path)

    if aoi.crs.to_epsg() != 4326:
        aoi = aoi.to_crs(epsg=4326)

    if not aoi.intersects(clipick_extent).any():
        raise Exception("Area of interest not covered by Clipick")
    
    if not aoi.within(clipick_extent).all():
        intersection_area = gpd.overlay(aoi, clipick_extent_gdf, how="intersection").geometry.area
        aoi_covered = round((100*intersection_area/aoi.area).values[0], 2)
        logging.warning(f" Only {aoi_covered}% of AoI covered by Clipick")
        aoi = gpd.clip(aoi, clipick_extent)

    create_sites_file(aoi, sites_path, start_date, end_date, step)
    get_data_from_clipick(sites_path, weather_path)


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='Create a grid of points')
    parser.add_argument('--aoi', type=str, required=True,
                        help='file with AOI')
    parser.add_argument('--sites_path', type=str, required=True,
                        help='path to where file with sites will be created')
    parser.add_argument('--weather_path', type=str, required=True,
                        help='path to where file with weather will be created')
    parser.add_argument('--step', type=float, default=25,
                        help='density of the grid in kilometers, defaults to 25')
    parser.add_argument('--start_date', type=str, required=True,
                        help='Starting date, either in format dd/mm/yyyy or just year (1/1)')
    parser.add_argument('--end_date', type=str, required=True,
                        help='End date, either in format dd/mm/yyyy or just year (31/12)')
    args = parser.parse_args()

    get_weather_data(args.aoi,
         args.sites_path,
         args.weather_path,
         args.start_date,
         args.end_date,
         args.step)




"""
Name: Generate a point grid within polygon
Author: Jan Pisl
Description: Generate a grid of points within
a polygon given as shapefile or similar. Generates
also points that are within a buffer of the polygon.
The points have an approximately equal distance in kilometers.

Conversion between kilometers and lat/lon are done using
the following formulas:
Latitude: 1 deg = 110.574 km
Longitude: 1 deg = 111.320*cos(latitude) km
https://stackoverflow.com/questions/1253499/simple-calculations-for-working-with-lat-lon-and-km-distance#:~:text=Latitude%3A%201%20deg%20%3D%20110.574%20km


Example of use:

python sites.py --shape test_aoi.shp --output_path sites_test.csv --start_date 2014 --end_date 2014
"""

import logging
import math
import csv
import argparse

import utm
import numpy as np
import geopandas as gpd
from shapely.geometry import Point


lat_conversion_const = 110.574
lon_conversion_const = 111.320


def create_grid(gdf, step=25):

    if not gdf.crs.to_epsg() == 4326:
        gdf = gdf.to_crs(epsg=4326)

    #TODO: handle multiple geometries
    if gdf.shape[0] > 1:
        raise NotImplementedError("Multiple shapes not handled yet.")

    # add a buffer of 2 steps in latitudial direction.
    # not perfect but likely good enough
    buffered_shape = gdf.loc[0].geometry.buffer(2*step/lat_conversion_const)

    min_lon, min_lat, max_lon, max_lat = buffered_shape.bounds

    lat, lon = min_lat, min_lon

    coords = []
    
    while lat < max_lat:
        while lon < max_lon:
            lon += step/(lon_conversion_const*math.cos(lat*2*math.pi/360))
            coords.append((lon, lat))
            
        lat += step/lat_conversion_const
        lon = min_lon

    filtered = [i for i in coords if buffered_shape.contains(Point(i))]

    return filtered


def get_headers(data):
    columns = len(data[0])
    if columns == 2:
        headers = ['lon', 'lat']
    elif columns == 8:
        headers = ['lon', 'lat', 'sy', 'sm', 'sd', 'ey', 'em', 'ed']
    else:
        raise Exception("Failed to determine what headers to use")

    return headers


def write_points_to_csv(grid, out_path):
    with open(out_path, "w", newline="") as f:
        writer = csv.writer(f)
        headers = get_headers(grid)
        writer.writerow(headers)
        writer.writerows(grid)

    logging.info(f"{len(grid)} points written to {out_path}")


def parse_date(date):
    if "/" in str(date):
        try:
            day, month, year = date.split("/")
        except ValueError:
            raise Exception(f"Invalid date: {date}")
    else:
        day, month, year = None, None, date

    return year, month, day


def add_dates(point_list, start_date, end_date):
    start_date = parse_date(start_date)
    if start_date[2] is None:
        start_date = (start_date[0], 1, 1)
    
    end_date = parse_date(end_date)
    if end_date[2] is None:
        end_date = (end_date[0], 12, 31)

    dated_points = [(*point, *start_date, *end_date) for point in point_list]

    return dated_points


def create_sites_file(aoi, out_path, start_date, end_date, step):

    grid_points = create_grid(aoi, step)

    points_with_dates = add_dates(grid_points, start_date, end_date)

    write_points_to_csv(points_with_dates, out_path)


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='Create a grid of points')
    parser.add_argument('--shape', type=str,
                        help='file with AOI')
    parser.add_argument('--output_path', type=str,
                        help='path to file with sites that will be created')
    parser.add_argument('--step', type=float, default=25,
                        help='density of the grid in kilometers, defaults to 25')
    parser.add_argument('--start_date', type=str,
                        help='Starting date, either in format dd/mm/yyyy or just year (1/1)')
    parser.add_argument('--end_date', type=str,
                        help='End date, either in format dd/mm/yyyy or just year (31/12)')
    args = parser.parse_args()

    create_sites_file(gpd.read_file(args.shape), args.step, args.start_date, args.end_date, args.output_path)




# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['prebas_weather']

package_data = \
{'': ['*']}

install_requires = \
['geopandas>=0.9.0,<0.10.0', 'requests>=2.25.1,<3.0.0', 'utm>=0.7.0,<0.8.0']

setup_kwargs = {
    'name': 'prebas-weather',
    'version': '0.1.0',
    'description': '',
    'long_description': None,
    'author': 'Jan Pisl',
    'author_email': 'jan.pisl@simosol.fi',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python

import pdb
import csv
import logging
import argparse
import requests

import pandas as pd 
import numpy as np 

format   = 'csv' # fmt
timespan = 'd'   # tspan
dataset  = 'METO-HC_HadRM3Q0_A1B_HadCM3Q0_DM_25km' # dts
asessment_report = 4 # ar
mode    = 'hisafe'   # mod
url     = 'http://www.isa.ulisboa.pt/proj/clipick/climaterequest_fast.php' 

CO2= 380

logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)



def transform_row(row_raw):
    row = [float(i) for i in row_raw.split(",")]
    PAR = row[7]*0.4278*4.56
    TAir = (row[3] + row[4])/2
    SVP = 610.7*10**(7.5*TAir/(237.3+TAir))
    VPD = SVP*(1-(row[5]+row[6])/2/100)/1000
    Precip = row[8]

    return PAR, TAir, VPD, Precip


string_date = {
    'sy': str, 
    'sm': int,
    'sd': int,
    'ey': str,
    'em': int,
    'ed': int
}


def make_request(df):

    response = requests.get(url, params={
        'lat': df.lat,
        'lon': df.lon,
        'fmt': format,
        'sy': df.sy,
        'sm': df.sm,
        'sd': df.sd,
        'ey': df.ey,
        'em': df.em,
        'ed': df.ed,
        'tspan': timespan,
        'dts': dataset,
        'ar': asessment_report,
        'mod': mode
    })

    return response


def get_data_from_clipick(sites, output_path):

    with open(output_path, 'w') as sink:
        sink_writer = csv.writer(sink)
        sink_writer.writerow(["clim_id", "PAR", "TAir", "VPD", "Precip", "CO2"])

        not_working = []

        sites_df = pd.read_csv(sites, dtype=string_date)
        sites_count = sites_df.shape[0]

        for index, df in sites_df.iterrows():
            
            response = make_request(df)

            '''# FILENAME: 3.123,965.34_2019.07-2020.07.csv
            lat_lon    = f'{df.lat:.3f},{df.lon:.3f}'; 
            start_date = f'{df.sy.strip()}.{df.sm:02}.{df.sd:02}'
            end_date   = f'{df.ey.strip()}.{df.em:02}.{df.ed:02}'
            filename   = f'{lat_lon}__{start_date}-{end_date}'''
            
            items = response.iter_lines()

            #skip headers
            headers = next(items)

            if not headers.decode('utf-8').lower() == 'day,month,year,tasmax,tasmin,hursmax,hursmin,rss,pr,wss':
                not_working.append(index)
                continue

            for row in items:
                row = row.decode('utf-8')

                #skip empty entries (what is the cause?)
                if row == '':
                    continue

                PAR, TAir, VPD, Precip = transform_row(row)
                sink_writer.writerow([index+1, PAR, TAir, VPD, Precip, CO2])

            if (index + len(not_working)) % 100 == 0:
                logging.info(f"{round(100*(index + len(not_working))/sites_count)}% of data acquired.")


    logging.info(f"Weather data for {sites_count-len(not_working)} sites written to {output_path}")
    if not_working:
        logging.warning(f"Data could not be acquired for the following sites:\n{not_working}")  


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='Get weather data from Clipick')
    parser.add_argument('--sites', type=str,
                        help='a CSV file containing lat,lon and dates for each site')
    parser.add_argument('--output', type=str,
                        help='path where the output will be written')

    args = parser.parse_args()

    get_data_from_clipick(args.sites, args.output)

    
